<div class="row element d-flex justify-content-{{ position }} {{ hide }} {{ css_class_container }} {{ align }}" id="{{ name }}-container" style="{{ height }}">
<div class="{{ element_width }}">
{# This is where the element HTML is injected #}
{{ html | safe}}
</div>
</div>
