<audio src="{{ src }}" {% if autoplay %}autoplay{% endif %} {% if loop %}loop{% endif %} {% if controls %}controls{% endif %} id="{{ name }}" class="{{ css_class_element }}" controlslist="nodownload">Your browser does not support the audio element.</audio>
