<input type="password" class="form-control input-field {{ align }} {{ css_class_element }}" id="{{ name }}" placeholder="{{ placeholder }}"
    name="{{ name }}" {% if input %} value="{{ input }}" {% elif default %} value="{{ default }}" {% endif %} {% if disabled %}disabled{% endif %} style="{{ height }} {{ fontsize }}">
