<textarea class="form-control input-field {{ align }} {{ css_class_element }}" name="{{ name }}" id="{{ name }}" rows="{{ area_nrows }}"
    placeholder="{{ placeholder }}" {% if disabled %}disabled{% endif %}
    style="{{ height }} {{ fontsize }}">{% if input %}{{ input }}{% elif default %}{{ default }}{% endif %}</textarea>
